#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dtbca.h"

#define ZERO 0

Graph::Graph(int _node_num_max, int _edge_num_max) 
  : node_num(0), node_num_max(_node_num_max),
    edge_num(0), edge_num_max(_edge_num_max),
    first_tree(NULL),
    lb(0), lb0(0),
    tree_block(128), tree_node_block(256), tree_edge_block(256), Mbuf(NULL), K_max(1)
{
  nodes = new Node[node_num_max];
  edges = new Edge[edge_num_max];
  
  lpdgs = new double[edge_num_max];
  
  node_lpdgs = new double[node_num_max];
  tmp_node_lpdgs = new double[node_num_max];
  combined_lpdgs = new double[edge_num_max];  // Separately store sum of node + edge lpdgs.
  
  cached_node_bounds = new double[node_num_max];
  cached_edge_bounds = new double[edge_num_max];
  
  num_updates_node_bonus = new double[node_num_max];
  for (int i = 0; i < node_num_max; i++) num_updates_node_bonus[i] = 1;
  
  spanning_tree_edge_ids = new int[node_num_max - 1]; // trees have exactly N-1 edges
  kruskal_sets = new DisjointSets(node_num_max);  // data structure for spanning tree connectivity
  vals_and_indices.reserve(edge_num_max);
  
  edges_adjacent_to_node.resize(node_num_max);
  
  capacity_scaling_threshold = 1e-5;
}

Graph::~Graph()
{
  delete [] nodes;
  delete [] edges;
  if (Mbuf) delete [] Mbuf;
  while (first_tree)
    {
      if (first_tree->message_array) delete [] first_tree->message_array;
      first_tree = first_tree->next;
    }
  // DB -- Danny forgot to delete dynamic arrays
  delete [] lpdgs;
  delete [] spanning_tree_edge_ids;
  delete kruskal_sets;  
}

void Graph::FinalizeGraph()
{
  if (Mbuf) { printf("FinalizeGraph() can be called at most once!\n"); exit(1); }
  
  Node* i;
  Edge* e;
  int K_sum = 0;
  K_max = 1;
  for (i=nodes; i<nodes+node_num; i++)
    {
      if (K_max < i->K) K_max = i->K;
      K_sum += i->K;
    }
  for (e=edges; e<edges+edge_num; e++)
    {
      K_sum += e->i[0]->K + e->i[1]->K;
    }
  
  Mbuf = new double[4*K_max + K_sum];
  double* ptr = Mbuf + 4*K_max;
  memset(ptr, 0, K_sum*sizeof(double));
  for (i=nodes; i<nodes+node_num; i++)
    {
      i->theta0 = ptr;
      ptr += i->K;
      memcpy(i->theta0, i->D, i->K*sizeof(double));
    }
  for (e=edges; e<edges+edge_num; e++)
    {
      e->M[0] = ptr;
      ptr += e->i[1]->K;
      e->M[1] = ptr;
      ptr += e->i[0]->K;
    }
  
  weight0_nodes = weight0_edges = 1;
}

Graph::EdgeId Graph::GetEdge(NodeId _i, NodeId _j)
{
  assert(_i>=0 && _i<node_num && _j>=0 && _j<node_num && edge_num<edge_num_max);
  Node* i = nodes + _i;
  Node* j = nodes + _j;
  Edge* e;
  for (e=i->first[0]; e; e=e->next[0])
    {
      if (e->i[1] == j) return (int)(e-edges);
    }
  for (e=i->first[1]; e; e=e->next[1])
    {
      if (e->i[0] == i) return (int)(e-edges);
    }
  return -1;
}

Graph::NodeId Graph::AddNode(int K, double* D)
{
  assert(Mbuf == NULL);
  assert(node_num < node_num_max);
  Node* i = nodes + node_num;
  i->K = K;
  i->D = D;
  i->first[0] = i->first[1] = NULL;
  i->first_tree_node = NULL;
  return node_num ++;
}

Graph::EdgeId Graph::AddEdge(NodeId _i, NodeId _j, V_PTR V)
{
  assert(_i>=0 && _i<node_num && _j>=0 && _j<node_num && edge_num<edge_num_max);
  Edge* e = edges + edge_num;
  Node* i = e->i[0] = nodes + _i;
  Node* j = e->i[1] = nodes + _j;
  e->V = V;
  e->next[0] = i->first[0];
  i->first[0] = e;
  e->next[1] = j->first[1];
  j->first[1] = e;
  e->first_tree_edge = NULL;
  e->weight0 = 1;
  vals_and_indices.push_back(make_pair(0, 0));  // DT 10/30/2010
  return edge_num ++;
}

Graph::TreeId Graph::AddTree(NodeId root)
{
  if (!Mbuf) { printf("Call FinalizeGraph() before AddTree()!\n"); exit(1); }
  
  assert(root >= 0 && root<node_num);
  Node* r = nodes + root;
  Tree* t = tree_block.New();
  t->message_array = NULL;
  t->message_array_size = r->K;
  
  t->next = first_tree;
  if (first_tree) first_tree->prev = t;
  first_tree = t;
  t->prev = NULL;
  
  t->lb = 0;
  
  TreeNode* ti = tree_node_block.New();
  ti->t = t;
  ti->i = r;
  
  ti->next = r->first_tree_node;
  if (r->first_tree_node) r->first_tree_node->prev = ti;
  r->first_tree_node = ti;
  ti->prev = NULL;
  
  ti->parent = NULL;
  ti->sibling = NULL;
  ti->first_child = NULL;
  t->root = ti;
  t->message_array = NULL;
  return t;
}

void Graph::AddTreeEdge(Tree* t, EdgeId _e, bool main_edge)
{
  assert(t->message_array == NULL);
  assert(_e>=0 && _e<edge_num);
  
  Edge* e = edges + _e;
  TreeNode* ti_ends[2] = { e->i[0]->first_tree_node, e->i[1]->first_tree_node };
  int dir = 0;
  while ( 1 )
    {
      if (!ti_ends[dir]) { dir = 1-dir; while(ti_ends[dir]->t!=t) ti_ends[dir] = ti_ends[dir]->next; break; }
      if (ti_ends[dir]->t == t) break;
      ti_ends[dir] = ti_ends[dir]->next;
      dir = 1 - dir;
    }
  TreeNode* ti = ti_ends[dir];
  assert(ti->i == e->i[0] || ti->i == e->i[1]);
  assert(ti->t == t);
  
  TreeNode* tj = tree_node_block.New();
  TreeEdge* te = tree_edge_block.New();
  
  te->e = e;
  te->dir = 1-dir;
  te->next = e->first_tree_edge;
  
  if (e->first_tree_edge) e->first_tree_edge->prev = te;
  e->first_tree_edge = te;
  te->prev = NULL;
  
  tj->t = t;
  tj->i = e->i[1-dir];
  tj->next = tj->i->first_tree_node;
  if (tj->i->first_tree_node) tj->i->first_tree_node->prev = tj;
  tj->i->first_tree_node = tj;
  tj->prev = NULL;
  
  te->parent = ti;
  te->child = tj;
  
  tj->parent = te;
  tj->first_child = NULL;
  tj->sibling = ti->first_child;
  ti->first_child = tj;
  
  t->message_array_size += tj->i->K;
  
  if (true) {
    te->M_to_parent = e->M[te->dir];
    te->M_to_child  = e->M[1-te->dir];
    te->weight = e->weight0;
    e->weight0 = 0;
  }
  
  else
    {
      t->message_array_size += e->i[0]->K + e->i[1]->K;
      te->M_to_parent = NULL;
      te->M_to_child  = NULL;
      te->weight = 0;
    }
}


void Graph::FinalizeTree(Tree* t)
{
  TreeNode* ti;
  
  assert(t->message_array == NULL);
  
  ti = t->root;
  while ( 1 )
    {
      ti->theta = ti->i->theta0;
      if (ti->parent && !ti->parent->M_to_parent)
	{
	  int Ki = ti->i->K, Kj = ti->parent->parent->i->K;
	}
      if (ti->first_child) ti = ti->first_child;
      else
	{
	  while (!ti->sibling && ti->parent) ti = ti->parent->parent;
	  if (!ti->parent) break;
	  ti = ti->sibling;
	}
    }
  
  t->weight_nodes = t->weight_edges = 1;
}


void Graph::RemoveTree(Tree* t)
{
  t->weight_nodes = 0;
  t->weight_edges = 0;
  //AverageTree(t);
  
  TreeNode* ti = t->root;
  while ( 1 )
    {
      if (ti->first_child) { ti = ti->first_child; continue; }
      
      TreeEdge* te = ti->parent;
      TreeNode* sibling = ti->sibling;
      
      if (ti->prev) ti->prev->next = ti->next;
      else ti->i->first_tree_node = ti->next;
      if (ti->next) ti->next->prev = ti->prev;
      tree_node_block.Delete(ti);
      
      if (!te) break;
      
      te->parent->first_child = sibling;
      ti = (sibling) ? sibling : te->parent;
      
      if (te->prev) te->prev->next = te->next;
      else te->e->first_tree_edge = te->next;
      if (te->next) te->next->prev = te->prev;
      tree_edge_block.Delete(te);
    }
  
  if (t->message_array) delete [] t->message_array;
  if (t->prev) t->prev->next = t->next;
  else first_tree = t->next;
  if (t->next) t->next->prev = t->prev;
  tree_block.Delete(t);
}

inline double ComputeMin(double* D, int K, int& x_min)
{
  x_min = 0;
  double D_min = D[0];
  int x;
  for (x=1; x<K; x++)
    {
      if (D_min > D[x]) { D_min = D[x]; x_min = x; }
    }
  return D_min;
}

double Graph::ReparameterizeTree(Tree* t, RepTreeMethod method, double omega)
{
  TreeNode* ti;
  TreeNode* tj;
  TreeEdge* te;
  double E = 0;
  double tmp_lb;
  int xi, xj, Ki, Kj;
  
  // forward pass
  ti = t->root;
  
  while (ti->first_child) { ti->i->tmp_count = 1; ti = ti->first_child; }
  ti->i->tmp_count = 1;
  
  while ( (te=ti->parent) ) {
    
    // pass message
    tj = te->parent;
    Ki = ti->i->K;
    Kj = tj->i->K;
    tj->i->tmp_count += ti->i->tmp_count;
    for (xi=0; xi<Ki; xi++) ti->theta[xi] -= te->M_to_child[xi];
    for (xj=0; xj<Kj; xj++) tj->theta[xj] -= te->M_to_parent[xj];
    tmp_lb = DISTANCE_TRANSFORM(ti->theta, te->M_to_parent, te->e->V, Ki, Kj, 1, te->dir);

    E += tmp_lb;
    for (xj=0; xj<Kj; xj++) tj->theta[xj] += te->M_to_parent[xj];
    
    // move ti
    if (ti->sibling) {
      ti = ti->sibling;

      while (ti->first_child) { ti->i->tmp_count = 1; ti = ti->first_child; }
      ti->i->tmp_count = 1;
    }
    else ti = ti->parent->parent;
  }
  
  // backward pass
  ti = t->root;
  ti->i->dirty = true;
  ti->i->lb_dirty = true;
  
  tmp_lb = ComputeMin(ti->theta, Ki, ti->i->x);
  E += tmp_lb;
  
  while (1) {
    Ki = ti->i->K;
    
    if ( method == METHOD_VNK ) {
      ////////////////////////////////////////////////////////////////////////////
      double alpha_remaining = 1;
      for (tj=ti->first_child; tj; tj=tj->sibling) {
	tj->i->dirty = true;
	tj->i->lb_dirty = true;
	Kj = tj->i->K;
	te = tj->parent;
	COMPUTE_MIN(tj->i->x, tj->theta, ti->i->x, te->e->V, Ki, Kj, 1, te->dir);
	double alpha = omega * tj->i->tmp_count / (double)ti->i->tmp_count;
	alpha_remaining -= alpha;
	
	for (xi=0; xi<Ki; xi++) {
	  te->M_to_parent[xi] -= alpha*ti->theta[xi];
	  Mbuf[xi] = -te->M_to_parent[xi];
	}
	DISTANCE_TRANSFORM(Mbuf, te->M_to_child, te->e->V, Ki, Kj, 1, 1-te->dir);
	for (xj=0; xj<Kj; xj++) tj->theta[xj] += te->M_to_child[xj];
      }
      for (xi=0; xi<Ki; xi++) ti->theta[xi] *= alpha_remaining;
      
      ////////////////////////////////////////////////////////////////////////////
      
    } else {
      
      // Sontag update
      for (tj = ti->first_child; tj; tj = tj->sibling) {
	tj->i->dirty = true;
	tj->i->lb_dirty = true;
	Kj = tj->i->K;
	te = tj->parent;
	COMPUTE_MIN(tj->i->x, tj->theta, ti->i->x, te->e->V, Ki, Kj, 1, te->dir);
	double alpha = tj->i->tmp_count / (double)ti->i->tmp_count;

	ti->i->tmp_count -= tj->i->tmp_count;
	
	for (xi=0; xi<Ki; xi++) {
	  Mbuf[xi] = ti->theta[xi] - te->M_to_parent[xi];
	  
	  te->M_to_parent[xi] -= alpha*ti->theta[xi];
	  ti->theta[xi]       -= alpha*ti->theta[xi];
	}
	
	DISTANCE_TRANSFORM(Mbuf, te->M_to_child, te->e->V, Ki, Kj, 1, 1-te->dir);
	
	for (xj=0; xj<Kj; xj++) {
	  te->M_to_child[xj] = (alpha-1)*tj->theta[xj] + alpha*te->M_to_child[xj];
	  tj->theta[xj] += te->M_to_child[xj];
	}
      }   
    }
    
    if (ti->first_child) ti = ti->first_child;
    else {
      while (!ti->sibling && ti->parent) ti = ti->parent->parent;
      if (!ti->parent) break;
      ti = ti->sibling;
    }
  }
  
  t->lb = E;
  
  return E;
}


void Graph::AverageEdge(Edge* e, bool compute_lower_bound)
{
  TreeEdge* te;
  double* M[2] = { Mbuf, Mbuf + K_max };
  int x;
  int K[2] = { e->i[0]->K, e->i[1]->K };
  double w_sum = 0;
  
  if (compute_lower_bound && !e->i[0]->lb_dirty && !e->i[1]->lb_dirty) {
    lb += cached_edge_bounds[e-edges];
    lb0 += cached_edge_bounds[e-edges];
    return;
  }
  
  for (x=0; x<K[0]; x++) M[1][x] = e->M[1][x];
  for (x=0; x<K[1]; x++) M[0][x] = e->M[0][x];
  w_sum = 1;
  
  if (w_sum <= 0) { printf("ERROR: each edge must be covered with non-zero weight (AverageEdge)!\n"); exit(1); }
  
  double v_min = 0;
  if (compute_lower_bound)
    {
      double* Di = Mbuf+2*K_max;
      double* Dj = Mbuf+3*K_max;
      for (x=0; x<K[0]; x++) Di[x] = -M[1][x];
      v_min = DISTANCE_TRANSFORM(Di, Dj, e->V, K[0], K[1], 1, 0, false);
      for (x=0; x<K[1]; x++) Dj[x] -= M[0][x];
      v_min += ComputeMin(Dj, K[1], x);
    }
  
  if (compute_lower_bound) {
    lb += v_min;
    lb0 += v_min;
    cached_edge_bounds[e-edges] = v_min;
  }
}

void Graph::AverageNode(Node* i, bool compute_lower_bound)
{
  Edge* e;
  TreeEdge* te;
  int dir;
  int x, K = i->K;
  double* theta = Mbuf;
  
  double v_min = 0;
  if (compute_lower_bound) {
    if (i->lb_dirty) {
      v_min = ComputeMin(i->theta0, K, x);
      cached_node_bounds[i-nodes] = v_min;
    } else {
      v_min = cached_node_bounds[i-nodes];
    }
    lb += v_min;
    lb0 += v_min;
  }
}

void Graph::AverageTree(Tree* t, bool compute_lower_bound)
{
  TreeNode* ti;
  TreeEdge* te;
  
  if (compute_lower_bound) ClearTreeLowerBound(t);
  
  ti = t->root;
  while (ti->first_child) ti = ti->first_child;
  while ( (te=ti->parent) )
    {
      AverageNode(ti->i, compute_lower_bound);
      AverageEdge(te->e, compute_lower_bound);
      
      // move ti
      if (ti->sibling)
	{
	  ti = ti->sibling;
	  while (ti->first_child) { ti->i->tmp_count = 1; ti = ti->first_child; }
	  ti->i->tmp_count = 1;
	}
      else ti = ti->parent->parent;
    }
  AverageNode(t->root->i, compute_lower_bound);
}

void Graph::AverageAll(bool compute_lower_bound)
{
  Node* i;
  Edge* e;
  
  if (compute_lower_bound)
    {
      ClearLowerBound();
      ClearLowerBound0();
      Tree* t;
      for (t=first_tree; t; t=t->next) ClearTreeLowerBound(t);
    }
  
  for (i=nodes; i<nodes+node_num; i++)
    {
      AverageNode(i, compute_lower_bound);
    }
  for (e=edges; e<edges+edge_num; e++)
    {
      AverageEdge(e, compute_lower_bound);
    }
}

double Graph::ComputeEnergy(int* labeling)
{
  Node* i;
  Edge* e;
  double E = 0;
  
  if (labeling)
    {
      int _i;
      for (i=nodes, _i=0; _i<node_num; i++, _i++) {
	//cout << "node contribution: " << i->D[labeling[_i]] << endl;
	E += i->D[labeling[_i]];
      }
      for (e=edges; e<edges+edge_num; e++) {
	//cout << "edge contribution: " << GET_COST(labeling[(int)(e->i[0]-nodes)], labeling[(int)(e->i[1]-nodes)], e->V, e->i[0]->K, e->i[1]->K) << endl;
	E += GET_COST(labeling[(int)(e->i[0]-nodes)], labeling[(int)(e->i[1]-nodes)], e->V, e->i[0]->K, e->i[1]->K);
      }
    }
  else
    {
      for (i=nodes; i<nodes+node_num; i++) {
	//cout << "node contribution: " << i->D[i->x] << endl;
	E += i->D[i->x];
      }
      for (e=edges; e<edges+edge_num; e++) {
	//cout << "edge contribution: " << GET_COST(e->i[0]->x, e->i[1]->x, e->V, e->i[0]->K, e->i[1]->K) << endl;
	E += GET_COST(e->i[0]->x, e->i[1]->x, e->V, e->i[0]->K, e->i[1]->K);
      }
    }
  
  return E;
}


int Graph::ComputeNodeLocalAssignment(int node_id, bool first_iteration) {
  
  Node *i = nodes+node_id;
  int x, K = i->K;
  
  if (first_iteration) {
    ComputeMin(i->D, K, x);
  } else {
    ComputeMin(i->theta0, K, x);
  }
  return x;
}

void Graph::ComputeEdgeWTAMeasures() {
  for (int edge_id = 0; edge_id < edge_num; edge_id++) {
    Edge *e = edges + edge_id;
    
    if (!e->i[0]->dirty && !e->i[1]->dirty)  continue;  // Don't recompute if no endpt is dirty
    
    // If the edge is not in some tree, give it a little boost of epsion + its lpdg
    // In KruskalNoZeros, we check to see if value is <= 1e-10, so this won't cause
    // us to add edges that really have zero lpdg.
    lpdgs[edge_id] = ComputeEdgeWTAMeasure(e) + ZERO;
    
  }    
}


void Graph::ComputeNodeWTAMeasures() {
  for (int node_id = 0; node_id < node_num; node_id++) {
    Node *n = nodes+node_id;
    
    if (!n->dirty)  continue;  // Don't recompute if not dirty
    
    node_lpdgs[node_id] = ComputeNodeWTAMeasure(n);
  }  
}


void Graph::MakeAllDirty() {
  for (int i = 0; i < node_num; i++) {
    (nodes+i)->dirty = true;
  }
}

void Graph::MakeAllClean() {
  for (int i = 0; i < node_num; i++) {
    (nodes+i)->dirty = false;
  }
}

void Graph::MakeAllLBClean() {
  for (int i = 0; i < node_num; i++) {
    (nodes+i)->lb_dirty = false;    
  }
}
void Graph::MakeAllLBDirty() {
  for (int i = 0; i < node_num; i++) {
    (nodes+i)->lb_dirty = true;    
  }
}

void Graph::StoreLabelingInHistories(int *labeling, int max_history_size, bool is_first_iteration) {
  int num_histories_stored = 0;
  for (int i = 0; i < node_num; i++) {
    Node *n = nodes+i;
    
    // Only store the label history if the variable is used in some tree.
    if (is_first_iteration || n->first_tree_node != NULL) {
      if (is_first_iteration) {
	n->label_history.clear();
      }
      if (n->label_history.size() >= max_history_size) {
	n->label_history.pop_back();
      }
      
      if (labeling == NULL) {
	n->label_history.push_front(n->x);
      } else {
	n->label_history.push_front(labeling[i]);
      }
      
      // Update label_set, which is just the unique elements in label_history
      n->label_set.erase(n->label_set.begin(), n->label_set.end());
      n->label_set.insert(n->label_history.begin(), n->label_history.end());
      
      num_histories_stored++;
    }
  }
}

void Graph::DilateLPDGs(int amt, bool is_grid) {
  for (int i = 0; i < node_num; i++) tmp_node_lpdgs[i] = 0;

  for (int eid = 0; eid < edge_num; eid++) {
    Edge *e = edges + eid;

    tmp_node_lpdgs[e->i[0]-nodes] += lpdgs[eid];
    tmp_node_lpdgs[e->i[1]-nodes] += lpdgs[eid];
  }
  
  if (is_grid) {
    for (int eid = 0; eid < edge_num; eid++) {
      Edge *e = eid+edges;
      int n0 = e->i[0]-nodes;
      int n1 = e->i[1]-nodes;
      
      double coin1 = rand() / static_cast<double>(RAND_MAX);
      double coin2 = rand() / static_cast<double>(RAND_MAX);
      if (coin1 < .05) {  // This is a parameter that seems to work well, but is heuristic
	if (lpdgs[eid] > capacity_scaling_threshold 
	    || node_lpdgs[n0] > capacity_scaling_threshold 
	    || node_lpdgs[n1] > capacity_scaling_threshold) {
	  
	  // decide whether to go out left/right or up/down
	  int delta = (coin2 < .5) ? 1 : size_x;
	  
	  int min_n = n0 < n1 ? n0 : n1;
	  int max_n = n0 > n1 ? n0 : n1;
	  
	  for (int d = 0; d < amt; d++) {
	    int new_n = max_n + d*delta;
	    if (new_n < node_num) {
	      tmp_node_lpdgs[new_n] += capacity_scaling_threshold + 1e-5;
	    }
	    new_n = min_n - d*delta;
	    if (new_n > 0) {
	      tmp_node_lpdgs[new_n] += capacity_scaling_threshold + 1e-5;
	    }
	  }
	}
      }
    }

  } else {  // Works on a general graph.. walk outward from chosen nodes
  
    bool marked[node_num];
    for (int eid = 0; eid < edge_num; eid++) {
      Edge *e = eid+edges;
      int n0 = e->i[0]-nodes;
      int n1 = e->i[1]-nodes;
      
      double coin = rand() / static_cast<double>(RAND_MAX);
      if ((lpdgs[eid] > capacity_scaling_threshold 
	   || node_lpdgs[n0] > capacity_scaling_threshold 
	   || node_lpdgs[n1] > capacity_scaling_threshold) && coin < .01) {
	
	// init
	for (int k = 0; k < node_num; k++) marked[k] = false;
	int cur_node = e->i[rand() % 2]-nodes;
	marked[cur_node] = true;
	
	for (int a = 0; a < amt; a++) {
	  // find num unmarked neighbors
	  int num_unmarked_neighbors = 0;
	  Node *n = nodes+cur_node;
	  for (int dir=0; dir<2; dir++) {
	    for (Edge *e=n->first[dir]; e; e=e->next[dir]) {
	      int other_node_id = e->i[1-dir]-nodes;
	      if (!marked[other_node_id]) {
		num_unmarked_neighbors++;
	      }
	    }
	  }
	  
	  if (num_unmarked_neighbors == 0)  break;
	  
	  // move to an unmarked neighbor
	  int neighbor_to_choose = rand() % num_unmarked_neighbors;
	  for (int dir=0; dir<2; dir++) {
	    for (Edge *e=n->first[dir]; e; e=e->next[dir]) {      
	      int other_node_id = e->i[1-dir]-nodes;
	      if (!marked[other_node_id]) {
		if (neighbor_to_choose == 0) {
		  marked[other_node_id] = true;
		  tmp_node_lpdgs[other_node_id] = capacity_scaling_threshold + 1e-5;
		  cur_node = other_node_id;
		}
		neighbor_to_choose--;
	      }
	    }
	  }
	}
      }
    }
  }
}


void Graph::ComputeCombinedLPDGs() {
  // Assumes we've called ComputeEdgeLPDGs and ComputeNodeLPDGs
  for (int edge_id = 0; edge_id < edge_num; edge_id++) {
    Edge *e = edges+edge_id;

    combined_lpdgs[edge_id] = lpdgs[edge_id] + node_lpdgs[e->i[0]-nodes] + node_lpdgs[e->i[1]-nodes];
  }
}

void Graph::AddDilatedLPDGs() {
  for (int edge_id = 0; edge_id < edge_num; edge_id++) {
    Edge *e = edges+edge_id;
    combined_lpdgs[edge_id] += tmp_node_lpdgs[e->i[0]-nodes] + tmp_node_lpdgs[e->i[1]-nodes];
  }
}

double Graph::ComputeEdgeWTAMeasure(Edge *e) {
  
  bool USE_ORIGINAL_POTENTIALS = false;  // an option just for debugging label sets.  this should be false in general.
  
  TreeEdge* te;
  double* M[2] = { Mbuf, Mbuf + K_max };
  int x;
  int K[2] = { e->i[0]->K, e->i[1]->K };
  double w_sum = 0;
  
  if (USE_ORIGINAL_POTENTIALS) {
    for (x=0; x<K[0]; x++) M[1][x] = 0;
    for (x=0; x<K[1]; x++) M[0][x] = 0;
  } else {
    for (x=0; x<K[0]; x++) M[1][x] = e->M[1][x];
    for (x=0; x<K[1]; x++) M[0][x] = e->M[0][x];
  }
  w_sum = 1;
  
  if (w_sum <= 0) { printf("ERROR: each edge must be covered with non-zero weight (EdgeWTAMeasure) !\n"); exit(1); }
  
  double v_min = 0;
  double* Di = Mbuf+2*K_max;
  double* Dj = Mbuf+3*K_max;
  for (x=0; x<K[0]; x++) Di[x] = -M[1][x];
  v_min = DISTANCE_TRANSFORM(Di, Dj, e->V, K[0], K[1], 1, 0);
  for (x=0; x<K[1]; x++) Dj[x] -= M[0][x];
  v_min += ComputeMin(Dj, K[1], x);
  
  double cost;
  Node *n0 = e->i[0];
  Node *n1 = e->i[1];
  set<int> &L0 = n0->label_set;
  set<int> &L1 = n1->label_set;
  set<int>::const_iterator it0;
  set<int>::const_iterator it1;
  double cur_cost;
  
  if (L0.size() == 0 || L1.size() == 0)  return 1000;
  
  // This is just to check that the distance transform call computes the minimum that we're
  // expecting it to.  This code can be removed later.
  bool debug_min = false;
  if (debug_min) {
    double all_min = HUGE_VAL;
    for (int i = 0; i < K[0]; i++) {
      cout << i << ":\t\t";
      for (int j = 0; j < K[1]; j++) {
	double cost = GET_COST(i, j, e->V, e->i[0]->K, e->i[1]->K) - M[1][i] - M[0][j];
	cout << cost << "\t";
	if (cost < all_min)  all_min = cost;
      }
      cout << endl;
    }
    CHECK_EPS_EQ(all_min, v_min, 1e-5, "Not calculating minimums properly");
  }
  
  // w_01 direction
  double worst_outer_cost_01 = -HUGE_VAL;
  for (it0 = L0.begin(); it0 != L0.end(); ++it0) {
    double best_inner_cost = HUGE_VAL;
    for (it1 = L1.begin(); it1 != L1.end(); ++it1) {
      cur_cost = GET_COST(*it0, *it1, e->V, e->i[0]->K, e->i[1]->K) - M[1][*it0] - M[0][*it1];
      if (cur_cost < best_inner_cost)  best_inner_cost = cur_cost;
    }
    if (best_inner_cost > worst_outer_cost_01)  worst_outer_cost_01 = best_inner_cost;
  }
  
  // w_10 direction
  double worst_outer_cost_10 = -HUGE_VAL;
  for (it1 = L1.begin(); it1 != L1.end(); ++it1) {
    double best_inner_cost = HUGE_VAL;
    for (it0 = L0.begin(); it0 != L0.end(); ++it0) {
      cur_cost = GET_COST(*it0, *it1, e->V, e->i[0]->K, e->i[1]->K) - M[1][*it0] - M[0][*it1];
      if (cur_cost < best_inner_cost)  best_inner_cost = cur_cost;
    }
    if (best_inner_cost > worst_outer_cost_10)  worst_outer_cost_10 = best_inner_cost;
  }
  
  double restricted_set_cost;
  if (worst_outer_cost_10 > worst_outer_cost_01) {
    restricted_set_cost = worst_outer_cost_10;
  } else {
    restricted_set_cost = worst_outer_cost_01;
  }
  
  double result = restricted_set_cost - v_min;
  
  return result;
}


double Graph::ComputeNodeWTAMeasure(Node *i) {
  Edge* e;
  int dir;
  int x, K = i->K;
  double* theta = Mbuf;
  
  memcpy(theta, i->theta0, K*sizeof(double));
  double v_min = 0;
  v_min = ComputeMin(theta, K, x);
  
  double cost;
  double worst_cost = -HUGE_VAL;
  set<int> &L_i = i->label_set;
  
  if (L_i.size() == 0) return 1000;  // first iteration, want to choose everything
  
  for (set<int>::iterator it = L_i.begin(); it != L_i.end(); ++it) {
    cost = theta[*it];
    if (cost > worst_cost)  worst_cost = cost;
  }
  
  CHECK_GT(L_i.size(), 0, "Empty label set!");
  
  return worst_cost - v_min;
}


// May want to let the choice of tree depend on the iteration; for example, if we're
// alternating horizontal and vertical combs.
// To compute LPDG based on different labeling, pass in optional last arg.
// Stores chosen edges in spanning_tree_edge_indices.
void Graph::ChooseTree(schedule_t schedule, 
		       int iteration, int *labeling) 
{

  if (schedule == COMBS_SCHEDULE || 
       (schedule != STARS_SCHEDULE && iteration < NUM_ITERATIONS_OF_COMBS)) 
  {
    kruskal_sets->ResetMembership();
    for (int i = 0; i < edges_adjacent_to_node.size(); i++) {
      edges_adjacent_to_node[i].clear();
    }
    
    if (iteration % 2 == 0) {
      ChooseSpanningTree_HorizontalComb();
    } else {
      ChooseSpanningTree_VerticalComb();
    }
    edges_are_in_topological_order = true;

  } 
  else if (schedule == STARS_SCHEDULE) 
  {
       
    kruskal_sets->ResetMembership();
    for (int i = 0; i < edges_adjacent_to_node.size(); i++) {
      edges_adjacent_to_node[i].clear();
    }
    
    ChooseSpanningTree_Star(iteration % node_num);
    edges_are_in_topological_order = true;
  
  } 
  else if (schedule == DYNAMIC_SCHEDULE) 
  {
    kruskal_sets->ResetMembership();
    for (int i = 0; i < edges_adjacent_to_node.size(); i++) {
      edges_adjacent_to_node[i].clear();
    }
    
    ComputeNodeWTAMeasures();

    ComputeEdgeWTAMeasures();
    
    ComputeCombinedLPDGs();      

    if (DILATE) {
      DilateLPDGs(10);
      AddDilatedLPDGs();
    }
      
    ChooseSpanningTree_KruskalNoZeros(capacity_scaling_threshold);
    
    edges_are_in_topological_order = false;

  } 
  else 
  {
    CHECK(false, "Unimplemented spanning tree algorithm");
  }
  
}


void Graph::ChooseSpanningTree_Star(int center_node) {
  int dir;
  int edge_id;
  Edge *e;
  Node *i = nodes + center_node;
  
  kruskal_num_edges_used = 0;
  
  //cout << "Star centered at " << center_node << endl;
  
  for (dir=0; dir<2; dir++) {
    for (e=i->first[dir]; e; e=e->next[dir]) {
      edge_id = e - edges;
      
      edges_adjacent_to_node[e->i[0]-nodes].push_back(edge_id);
      edges_adjacent_to_node[e->i[1]-nodes].push_back(edge_id);
      
      //cout << "\t" << e->i[1-dir]-nodes << "\t" << e->i[0]-nodes << "--" << e->i[1]-nodes << endl;
      
      spanning_tree_edge_ids[kruskal_num_edges_used] = edge_id;
      
      kruskal_num_edges_used++;
    }
  }
}


void Graph::ChooseSpanningTree_HorizontalComb() {
  
  kruskal_num_edges_used = 0;
  int increment = size_x;
  for (int i = 0; i < edge_num; i++) {
    Edge *e = edges+i;
    int n0 = e->i[0]-nodes;
    int n1 = e->i[1]-nodes;
    
    if (n0 < increment || n1 < increment) {
      edges_adjacent_to_node[e->i[0]-nodes].push_back(i);
      edges_adjacent_to_node[e->i[1]-nodes].push_back(i);
      
      CHECK(kruskal_num_edges_used < node_num_max - 1, "Using too many edges in spanning tree");
      
      spanning_tree_edge_ids[kruskal_num_edges_used] = i;
      kruskal_num_edges_used++;
    }
  }
  
  for (int i = 0; i < edge_num; i++) {
    Edge *e = edges+i;
    int n0 = e->i[0]-nodes;
    int n1 = e->i[1]-nodes;
    
    if (n0 >= increment && n1 >= increment && abs(n0-n1) == increment) {
      edges_adjacent_to_node[e->i[0]-nodes].push_back(i);
      edges_adjacent_to_node[e->i[1]-nodes].push_back(i);
      
      CHECK(kruskal_num_edges_used < node_num_max - 1, "Using too many edges in spanning tree");
      
      spanning_tree_edge_ids[kruskal_num_edges_used] = i;
      kruskal_num_edges_used++;
    }
  }
}


void Graph::ChooseSpanningTree_VerticalComb() {
  
  kruskal_num_edges_used = 0;
  int increment = size_x;
  for (int i = 0; i < edge_num; i++) {
    Edge *e = edges+i;
    int n0 = e->i[0]-nodes;
    int n1 = e->i[1]-nodes;
    
    if (n0 % increment == 0 && n1 % increment == 0) {
      edges_adjacent_to_node[e->i[0]-nodes].push_back(i);
      edges_adjacent_to_node[e->i[1]-nodes].push_back(i);
      
      CHECK(kruskal_num_edges_used < node_num_max - 1, "Using too many edges in spanning tree");
      spanning_tree_edge_ids[kruskal_num_edges_used] = i;
      
      kruskal_num_edges_used++;
    }
  }
  for (int i = 0; i < edge_num; i++) {
    Edge *e = edges+i;
    int n0 = e->i[0]-nodes;
    int n1 = e->i[1]-nodes;
    
    if (!(n0 % increment == 0 && n1 % increment == 0) && abs(n0 - n1) == 1) {
      edges_adjacent_to_node[e->i[0]-nodes].push_back(i);
      edges_adjacent_to_node[e->i[1]-nodes].push_back(i);
      
      CHECK(kruskal_num_edges_used < node_num_max - 1, "Using too many edges in spanning tree");
      spanning_tree_edge_ids[kruskal_num_edges_used] = i;
      
      kruskal_num_edges_used++;
    }
  }
}


void Graph::ChooseSpanningTree_Kruskal() {
  for (uint e_id = 0; e_id < edge_num; e_id++) {
    vals_and_indices[e_id] = make_pair(lpdgs[e_id], e_id);
  }
  sort(vals_and_indices.begin(), vals_and_indices.end(), PairCompare);
  
  int num_edges_used = 0;
  for (int i = 0; i < edge_num; i++) {
    
    int edge_id = vals_and_indices[i].second;
    Edge *e = edges + edge_id;
    int node_id0 = e->i[0]-nodes;
    int node_id1 = e->i[1]-nodes;
    int endpt1_label = kruskal_sets->FindSet(node_id0);
    int endpt2_label = kruskal_sets->FindSet(node_id1);
    
    if (endpt1_label != endpt2_label) {
      kruskal_sets->Union(endpt1_label, endpt2_label);  // Merge labels
      spanning_tree_edge_ids[num_edges_used] = edge_id;  // Add edge to spanning tree
      
      // Need to maintain adjacency info for adding to tree later.
      edges_adjacent_to_node[node_id0].push_back(edge_id);
      edges_adjacent_to_node[node_id1].push_back(edge_id);
      
      num_edges_used++;
    }
    
    if (num_edges_used == node_num - 1)  break;
  }
  kruskal_num_edges_used = node_num - 1;
}


void Graph::ChooseSpanningTree_KruskalNoZeros(double thresh) {
  bool use_combined_lpdgs = true;
  
  int num_used = 0;
  
  for (uint e_id = 0; e_id < edge_num; e_id++) {
    if (use_combined_lpdgs) {
      
      if (combined_lpdgs[e_id] > ZERO + thresh) {
	vals_and_indices[num_used] = make_pair(combined_lpdgs[e_id], e_id);
	num_used++;
      }
    } else {
      vals_and_indices[e_id] = make_pair(lpdgs[e_id], e_id);
    }
  }
  
  // Run spanning tree algorithm
  sort(vals_and_indices.begin(), vals_and_indices.begin() + num_used, PairCompare);
  kruskal_num_edges_used = 0;
  for (int i = 0; i < num_used; i++) {
    // Stop adding edges if the edge measure if below thresh.  This leads to
    // a forest instead of a tree, and avoids wasteful message passing
    // on irrelevant regions of the graph.
    if (vals_and_indices[i].first <= ZERO + thresh)  break;
    
    int edge_id = vals_and_indices[i].second;
    Edge *e = edges + edge_id;
    int node_id0 = e->i[0]-nodes;
    int node_id1 = e->i[1]-nodes;
    int endpt1_label = kruskal_sets->FindSet(node_id0);
    int endpt2_label = kruskal_sets->FindSet(node_id1);
    
    if (endpt1_label != endpt2_label) {
      kruskal_sets->Union(endpt1_label, endpt2_label);  // Merge labels
      spanning_tree_edge_ids[kruskal_num_edges_used] = edge_id;  // Add edge to spanning tree
      
      // Need to maintain adjacency info for adding to tree later.
      edges_adjacent_to_node[node_id0].push_back(edge_id);
      edges_adjacent_to_node[node_id1].push_back(edge_id);
      
      kruskal_num_edges_used++;
    }
    
    if (kruskal_num_edges_used == node_num - 1)  break;
  }
}


void Graph::SerializeMessagesAndTheta0s(ofstream &out) {
  int xi, xj;
  Edge* e;
  
  // Setting this to be less precise gives slightly different dual value on save-then-load.
  // This makes large files, though (e.g., 118MB for tsukuba).
  out.precision(15);
  
  for (e=edges; e<edges+edge_num; e++) {
    int Ki = e->i[0]->K, Kj = e->i[1]->K;
    for (xj=0; xj<Kj; xj++) out << e->M[0][xj] << " ";
    out << endl;
    
    for (xi=0; xi<Ki; xi++) out << e->M[1][xi] << " ";
    out << endl;
  }
}


void Graph::LoadMessagesAndTheta0s(ifstream &in) {
  int xi, xj;
  Edge* e;
  int dir;
  Node *i;
  char endline;
  
  for (e=edges; e<edges+edge_num; e++) {
    int Ki = e->i[0]->K, Kj = e->i[1]->K;
    for (xj=0; xj<Kj; xj++) in >> e->M[0][xj];
    //in.get(endline);  // read newline not necessary
    
    for (xi=0; xi<Ki; xi++) in >> e->M[1][xi];
    //in.get(endline);  // read newline not necessary
  }
  
  // Need to recompute theta0's.  They aren't actually stored in the file.
  for (i=nodes; i<nodes+node_num; i++) {
    for (xi=0; xi<i->K; xi++) i->theta0[xi] = i->D[xi];
    for (dir=0; dir<2; dir++) {
      for (e=i->first[dir]; e; e=e->next[dir]) {
	for (xi=0; xi<i->K; xi++) i->theta0[xi] += e->M[1-dir][xi];
      }
    }
  }
}


void Graph::IncrementMaskUnaries(string filename, double *delta) {
  Node *n;
  
  ifstream in;
  //cout << "Opening " << filename << endl;
  in.open(filename.c_str());
  
  for (n=nodes; n<nodes+node_num; n++) {
    n->dirty = false;
  }
  
  int node_id;
  char endline;
  string line;
  while (!in.eof()) {
    getline(in, line);
    if (line == "") break;
    node_id = atoi(line.c_str());
    //cout << node_id << " ";
    
    n = nodes+node_id;
    n->dirty = true;
    for (int xi=0; xi < n->K; xi++) {
      n->D[xi] += delta[xi];
      n->theta0[xi] += delta[xi];
    }
  }
  //cout << endl;
}


void Graph::SetNodeLabel(int node_id, int label, int K) {
  Node *n = nodes+node_id;
  n->dirty = true;
  
  for (int xi=0; xi < n->K; xi++) {
    double delta = (xi == label) ? 0 : 1e8;
    n->D[xi] += delta;
    n->theta0[xi] += delta;
  }
}


void Graph::Print0()
{
  int xi, xj;
  Node* i;
  Edge* e;
  
  printf("graph0:\n");
  for (i=nodes; i<nodes+node_num; i++)
    {
      printf("  node %d:", (int)(i-nodes));
      for (xi=0; xi<i->K; xi++) printf(" %f", i->theta0[xi]);
      printf("\n");
    }
  for (e=edges; e<edges+edge_num; e++)
    {
      printf("  edge %d (%d - %d), w=%f", (int)(e-edges), (int)(e->i[0]-nodes), (int)(e->i[1]-nodes), e->weight0);
      if (e->te == NULL)
	{
	  int Ki = e->i[0]->K, Kj = e->i[1]->K;
	  printf(" M01=");
	  for (xj=0; xj<Kj; xj++) printf("%c%f", (xj==0)?'[':' ', e->M[0][xj]);
	  printf("] M10=");
	  for (xi=0; xi<Ki; xi++) printf("%c%f", (xi==0)?'[':' ', e->M[1][xi]);
	  printf("]\n");
	  for (xi=0; xi<Ki; xi++)
	    {
	      printf("   ");
	      for (xj=0; xj<Kj; xj++) printf(" %f", e->weight0*GET_COST(xi, xj, e->V, Ki, Kj)-e->M[0][xj]-e->M[1][xi]);
	      printf("\n");
	    }
	}
      else printf("\n");
    }
  printf("\n");
}

void Graph::PrintTree(Tree* t)
{
  int xi, xj;
  TreeNode* ti;
  TreeEdge* te;
  
  printf("tree, root:%d:\n", (int)(t->root->i-nodes));
  
  ti = t->root;
  while (ti->first_child) ti = ti->first_child;
  while ( 1 )
    {
      printf("  node %d:", (int)(ti->i-nodes));
      for (xi=0; xi<ti->i->K; xi++) printf(" %f", ti->theta[xi]);
      printf("\n");
      
      te = ti->parent;
      if (!te) break;
      
      Edge* e = te->e;
      printf("  edge %d (%d - %d), w=%f", (int)(e-edges), (int)(e->i[0]-nodes), (int)(e->i[1]-nodes), te->weight);
      int Ki = e->i[0]->K, Kj = e->i[1]->K;
      double* M[2];
      M[te->dir] = te->M_to_parent;
      M[1-te->dir] = te->M_to_child;
      printf(" M01=");
      for (xj=0; xj<Kj; xj++) printf("%c%f", (xj==0)?'[':' ', M[0][xj]);
      printf("] M10=");
      for (xi=0; xi<Ki; xi++) printf("%c%f", (xi==0)?'[':' ', M[1][xi]);
      printf("]\n");
      for (xi=0; xi<Ki; xi++)
	{
	  printf("   ");
	  for (xj=0; xj<Kj; xj++) printf(" %f", te->weight*GET_COST(xi, xj, e->V, Ki, Kj)-M[0][xj]-M[1][xi]);
	  printf("\n");
	}
      
      // move ti
      if (ti->sibling)
	{
	  ti = ti->sibling;
	  while (ti->first_child) { ti->i->tmp_count = 1; ti = ti->first_child; }
	  ti->i->tmp_count = 1;
	}
      else ti = ti->parent->parent;
    }
  printf("\n");
}

