//
//  dtbca_main.cpp
//  dtbca_public_release
//
//  Created by Dhruv Batra on 8/25/11.
//  Copyright 2011 Dhruv Batra. All rights reserved.

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <fstream>
#include <iostream>
#include <list>
#include <sys/stat.h>

#include "dtbca.h"
#include "trw_s.h"
#include "mplp.h"

#include "utils.h"
#include "utils.hpp"

#include "constants_declaration.h"

// Simple Class to keep track of time-breakup
class time_breakup {
public:
    time_breakup() : t_initial_labelling(0), t_spanning_tree(0), t_adding_tree(0), t_repavg(0) {}
    
    double t_initial_labelling; // Time to compute a primal solution (used to compute LPDG/WTA)
    double t_spanning_tree;  // Time to find a spanning tree
    double t_adding_tree; // Time to add that spanning tree to the graph
    double t_repavg; // Time to reparameterize tree(s)
    
};
inline std::ostream& operator<<(std::ostream& os, const time_breakup& x)
{
    os << x.t_initial_labelling << ' ' << x.t_spanning_tree	\
    << ' ' << x.t_adding_tree << ' ' << x.t_repavg << endl;
    return os;
}

void NormalizeArray(double* array, int num, int size)
{
    int i, k;
    for (i=0; i<num; i++)
    {
        double v_min = array[0];
        for (k=1; k<size; k++) if (v_min > array[k]) v_min = array[k];
        for (k=0; k<size; k++) { *array -= v_min; array ++; }
    }
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// MAIN workhorse function. Implements DTBCA Algorithm as Described in 
// Tarlow, Batra, Kohli, Kolmogorov, ICML '11. 
// 
// Input Arguments:
// alg              -- 0 : star-TBCA, 1 : comb-TBCA, 2 : LPDG-based TBCA, 3 : WTA(10)-based TBCA
// nnodes           -- number of nodes in MRF
// nedges           -- number of edges in MRF
// cardinalities    -- (pointer to) array of length nnodes holding no. of states each MRF var can take
// edgelist         -- (pointer to) array of length 2*nedges holding list of (i,j) pairs; e.g. edgelist[2*e] = i; edgelist[2*e+1] = j; Node numbering starts from 0.
// D                -- (pointer to) array holding node potentials; e.g. if all nodes have same no. of labels D[i*nlabels + l] = cost of node i taking state l (label numbering starts from 0).
// V                -- (pointer to) array holding edge potentials in *column-major* format; e.g. if all nodes have same no. of labels V[e*nlabels^2 + li + lj*nlabels] = cost of edge e taking labels (li,lj) (edge numbering starts from 0).
// sizeX            -- Optional argument. In case of a grid graph, set this to ncols.
// sizeY            -- Optional argument. In case of a grid graph, set this to nrows. 

void dtbca(int alg, int nnodes, int nedges, int *cardinalities, 
          int* edgelist, double* &D, double* &V, int sizeX, int sizeY) // Last two are optional arguments
{
    
    ////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////
    // Parameters
    // This code allows for saving and loading of messages. 
    bool LOAD_MESSAGES = false;
    bool SAVE_MESSAGES = false;
    bool MODIFY_MASK = false;
    
    schedule_t schedule;
    int WTA_HISTORY_SIZE = -1;
    switch (alg) 
    {
        case 0:
            schedule = STARS_SCHEDULE;
            break;
        case 1:
            schedule = COMBS_SCHEDULE;
            CHECK(sizeX*sizeY>0, "Combs schedule is only defined for grid-graph (nrows,ncols cannot be undefined)");
            break;
        case 2:
            schedule = DYNAMIC_SCHEDULE;    
            WTA_HISTORY_SIZE = 1;
            break;
        case 3:
            schedule = DYNAMIC_SCHEDULE;    
            WTA_HISTORY_SIZE = 10;
            break;
        default:
            CHECK(false, "Invalid alg");
            break;
    }
    
    
    ////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////
    // Graph Construction
    Graph *g = new Graph(nnodes,nedges);
    
    g->size_x = sizeX;
    g->size_y = sizeY;
    
    // Add node potentials
    int cardsum = 0;
    for (int i=0; i!=nnodes; ++i)
    {
        CHECK_NOT_NULL(D+cardsum, "Adding NULL unary potential");
        g->AddNode(cardinalities[i], D+cardsum);
        cardsum += cardinalities[i];
    }
    
    // Add edge potentials
    int ni, nj;
    cardsum = 0;
    for (int e=0; e!=nedges; ++e) 
        {
            ni = edgelist[2*e];
            nj = edgelist[2*e+1];
            

#ifdef TYPE_POTTS
            g->AddEdge(ni, nj, V[e]);
#else
            CHECK_NOT_NULL(V+cardsum, "Adding NULL pairwise potential");
            g->AddEdge(ni, nj, V+cardsum);
            cardsum += cardinalities[ni] * cardinalities[nj];
#endif
        }

    // Finalized Graph
    g->FinalizeGraph();

        
    if (LOAD_MESSAGES) {
        ifstream serialized_msgs_in;
        string serialized_filename = name + "/" + name + "_serialized_msgs.dat";
        serialized_msgs_in.open(serialized_filename.c_str());
        g->LoadMessagesAndTheta0s(serialized_msgs_in);
        serialized_msgs_in.close();
    }
    
    if (MODIFY_MASK) {
        double *delta_vector = new double[(g->nodes)->K];
        for (int i = 0; i < (g->nodes)->K; i++) {
            if (i == 0) {
                delta_vector[i] = 0;
            } else {
                delta_vector[i] = 10;
            }
        }
        g->IncrementMaskUnaries(mask_filename, delta_vector);
    }
    
    ////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////
    // Do message-passing
    
    // Tree Variables
    vector<Graph::TreeId> t;
    int num_trees = t.size(), num_new_trees = 0, num_tree_edges;

    // Timing Variables
    double tt, total_time = 0;
    MyTimer timer1, timer2;

    // Other book-keeping variables
    vector<double> lbhist, enhist,timehist, edgecnthist, disthist; // to make lb/en-vs-iter/time plots
    vector<time_breakup> tbhist;
    double lb, prev_lb;
    double cost = INF-1, best_cost = INF;

    int iter=0, k; 
    int label_buf[nnodes];
    int *initial_labeling;
    bool marked[nnodes];
    
    bool converged = false;    
    g->SetEdgeWeights(0);
    g->SetNodeWeights(0);
    g->MakeAllDirty();

    // Tree-Block Coordinate Ascent Iterations
    while (!converged)
    {
        // Time the entire iteration
        timer1.tic();

        time_breakup tb; 
        num_new_trees = 0;
        // keep track of num calls to dist trans in this iteration
        g->num_dist_trans = 0;
        
        // Need these on to compute LPDGs.
        g->SetEdgeWeights(1);
        g->SetNodeWeights(1);
        
        
        ////////////////////////////////////////////////////////////////////////////////
        // Compute a primal solution (Labelling)
        timer2.tic(); // START Initial Labeling
        
        initial_labeling = label_buf;
        // For now, we're always using node-decoded labeling
        for (int i = 0; i < nnodes; i++) 
            initial_labeling[i] = g->ComputeNodeLocalAssignment(i, false);
        
        if (iter == 0 || iter == NUM_ITERATIONS_OF_COMBS) 
            g->StoreLabelingInHistories(initial_labeling, WTA_HISTORY_SIZE, true);
        
        tb.t_initial_labelling += timer2.toc(false); // END Initial Labeling
        
        
        ////////////////////////////////////////////////////////////////////////////////
        // Find the Spanning Tree that uses most of LPDG/WTA scores
        timer2.tic(); // START Compute Spanning Tree
        g->ChooseTree(schedule, iter, initial_labeling);        
        tb.t_spanning_tree += timer2.toc(false); // END Compute Spanning Tree
        

        ////////////////////////////////////////////////////////////////////////////////
        // Add this Spanning Tree to the Graph Class
        // This is easy if the edges are in topogial order, 
        // but otherwise little cumbersome because of the code
        timer2.tic(); // START Add Spanning Tree        
        int *tree_edge_ids = g->SpanningTreeEdgeIndices();
        num_tree_edges = g->kruskal_num_edges_used;
        
        //cout << num_tree_edges << "\t";
        
        if (g->edges_are_in_topological_order) 
	    {
            int n0 = (g->edges+tree_edge_ids[0])->i[0]-g->nodes;
            t.push_back(g->AddTree(n0));
            
            for (int i = 0; i < g->kruskal_num_edges_used; i++) {
                g->AddTreeEdge(t.back(), tree_edge_ids[i]);
            }
            g->FinalizeTree(t.back());
            
	    } 
        else
	    {
            // NOTE: assumes separate trees are disjoint.  If they aren't, it will
            // try to add a big combined tree.  This could be a problem if the union
            // of two trees forms a cycle.  In our non-spanning trees formulation, though,
            // we'll always have disjoint trees.
            for (int i = 0; i < nnodes; i++) {
                marked[i] = false;
            }
            
            set<int> unmarked_nodes;
            int eid;
            int n0;
            int n1;	
            
            // Only want to insert nodes that are adjacent to some edge.
            for (int i = 0; i < g->kruskal_num_edges_used; i++) {
                eid = tree_edge_ids[i];		 
                n0 = (g->edges+eid)->i[0]-g->nodes;
                n1 = (g->edges+eid)->i[1]-g->nodes;
                
                unmarked_nodes.insert(n0);
                unmarked_nodes.insert(n1);
            }
            
            while (true) {  // Loop over trees
                bool added_new_tree = false;
                
                if (unmarked_nodes.empty()) break;  // We've added all edges
                n0 = *unmarked_nodes.begin();  // arbitrary first unmarked node in set
                
                //cout << "  Starting with root " << n0 << endl;
                //if (marked[n0])  cout << "ERROR: node is marked!" << endl;
                
                list<int> nodes_to_expand;
                nodes_to_expand.push_back(n0);
                
                int other_node;
                while (!nodes_to_expand.empty()) {
                    int cur_node = nodes_to_expand.front();
                    nodes_to_expand.pop_front();
                    
                    if (marked[cur_node]) continue;
                    
                    list<int>::const_iterator it;
                    for (it = g->edges_adjacent_to_node[cur_node].begin();
                         it != g->edges_adjacent_to_node[cur_node].end(); ++it) {
                        eid = *it;
                        
                        n0 = (g->edges+eid)->i[0]-g->nodes;
                        n1 = (g->edges+eid)->i[1]-g->nodes;
                        
                        if (!added_new_tree) {
                            t.push_back(g->AddTree(n0));
                            num_new_trees++;
                            num_trees++;
                            added_new_tree = true;
                        }
                        
                        if (n0 == cur_node) other_node = n1;
                        else other_node = n0;
                        
                        if (!marked[other_node]) {
                            g->AddTreeEdge(t.back(), eid);
                        }
                        
                        nodes_to_expand.push_back(other_node);
                    }
                    marked[cur_node] = true;
                    unmarked_nodes.erase(cur_node);
                }
                g->FinalizeTree(t.back());
            }
	    } 
                
        tb.t_adding_tree += timer2.toc(false); // END Add Spanning Tree
        
        
        // This is just for visualization.  In reality, we should set clean after computing
        // node measures.
        g->MakeAllClean();
        
        ////////////////////////////////////////////////////////////////////////////////
        // Reparameterize Trees based on strategy
        // Compute lower-bound and energy at the end of this TBCA step. 
        // Sontag09 TBCA keeps around a set of trees and operates over this mixture. 
        // Our code supports that, but currently functions by iteratively giving a 
        // a single tree full weight.
        
        timer2.tic();  // start rep-avg
        
        lb = 0;        
        for (int i=t.size()-1; i>=0; --i) {
            
            // Now reparameterize tree i and get lower-bound
            lb += g->ReparameterizeTree(t[i], Graph::METHOD_SONTAG);
            //lb += g->ReparameterizeTree(t[i], Graph::METHOD_VNK, .75);
	    }
	    
	    // Get lb from edges not covered by any tree	    
	    //g->SetEdgeWeights(1);	
	    //g->SetNodeWeights(1);	    
	    g->AverageAll(true);
	    lb = g->GetLowerBound0();
	    g->MakeAllLBClean();
	    
	    for (int i = 0; i < nnodes; i++) {
            initial_labeling[i] = g->ComputeNodeLocalAssignment(i, false);
	    }
	    
	    if (schedule == DYNAMIC_SCHEDULE) {
            if (iter > NUM_ITERATIONS_OF_COMBS) {
                g->StoreLabelingInHistories(initial_labeling, WTA_HISTORY_SIZE, iter == 0);  // DT 1/18
            }
	    }
	    
	    cost = g->ComputeEnergy(initial_labeling);
	    
	    // We can keep the current (set of) trees around for a few itearions 
        // (to reduce spanning tree computations).
        // But this version of the code clears and finds a new tree in each iteration. 
	    int t_size = t.size();
	    if (true || iter == 0 || lb > (prev_lb))
        {
            //g->SetEdgeWeights(1);	
            //g->SetNodeWeights(1);
            
            for (int i = 0; i < t.size(); i++)
                g->RemoveTree(t[i]);
            t.clear();
            num_trees = 0;
	    }
	    prev_lb = lb;
	    
        tb.t_repavg += timer2.toc(false); // END Reparameterize and Average

        tt = timer1.toc(false);
        total_time += tt;
        
        //g->Print0();
        printf("iter %d, time %lf (%f total): lower-bound = %f, energy = %f, #edges rep = %d\n", iter, tt, total_time, lb, cost, g->num_dist_trans);

 
        // keep track of lb/en/time
        lbhist.push_back(lb); enhist.push_back(cost);
        timehist.push_back(tt); tbhist.push_back(tb);
        edgecnthist.push_back(num_tree_edges);
        disthist.push_back(g->num_dist_trans);
        
        
        ////////////////////////////////////////////////////////////////////////////////
        // Check Convergence of Tree-Block Coordinate Ascent
        
        // break if integrality gap < thr;
        if (cost-lb < INT_GAP_THRES)  
            converged = true;
        // break if dual has converged
        if ( (iter>1) && (lb - lbhist[lbhist.size()-2] < DUAL_IMPROV_THRES) && (alg != 0) ) // stars only update one star per iter
                                                                                            // so only check for non-star updates
            converged = true;
        // break if enough iterations or time has passed
        if ( (iter > MAXITER) || (total_time > MAXTIME) )
            converged = true;

        iter++;
    }
        
    ////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////
    // Save data to files
    ostringstream o; o << "_alg_t" << alg << ".txt";

    WriteToFile(outputpath+"enhist"+o.str(), enhist);
    WriteToFile(outputpath+"lbhist"+o.str(), lbhist);
    WriteToFile(outputpath+"timehist"+o.str(), timehist);
    WriteToFile(outputpath+"edgecnthist"+o.str(), edgecnthist);
    WriteToFile(outputpath+"tbhist"+o.str(), tbhist);
    WriteToFile(outputpath+"disthist"+o.str(), disthist);
    WriteToFile(outputpath+"primalsol"+o.str(),vector<int>(label_buf, label_buf+nnodes));

    if (SAVE_MESSAGES) 
    {
        ofstream serialized_msg_file;
        string serialized_filename = name + "/" + name + "_serialized_msgs.dat";
        serialized_msg_file.open(serialized_filename.c_str());
        g->SerializeMessagesAndTheta0s(serialized_msg_file);
        serialized_msg_file.close();
    }
    
    ////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////
    // Destruction and Done
    delete [] D; D = NULL;
    delete [] V; V = NULL;
    delete g;
}

